from typing import Any, AsyncIterator
import pickle as pkl

import pandas as pd
import pyarrow as pa

from sarus_data_spec.manager.asyncio.utils import async_iter
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

from .pandas import pd_any, pd_eq, pd_loc, pd_mean, pd_std
from .sklearn import sk_fit
from .std import add, div, mul, sub


async def arrow_external(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    """Call external and convert the result to a RecordBatch iterator.

    We consider that external ops results are Datasets. For now, we consider
    that pandas.DataFrame are Datasets. For instance, the pd.loc operation only
    selects a subset of a Dataset and so is a Dataset.

    We call the implementation of `external` which returns arbitrary values,
    check that the result is indeed a DataFrame and convert it to a RecordBatch
    iterator.
    """
    val = await external(dataset)
    if isinstance(val, pd.DataFrame):
        return async_iter(
            pa.Table.from_pandas(val).to_batches(max_chunksize=batch_size)
        )

    else:
        raise TypeError(f"Cannot convert {type(val)} to Arrow batches.")


async def external(dataspec: st.DataSpec) -> Any:
    """Route an externally transformed Dataspec to its implementation."""
    transform_spec = dataspec.transform().protobuf().spec
    external_op = sp.Transform.ExternalOp.Name(transform_spec.external.op)
    implemented_ops = {
        "ADD": add,
        "MUL": mul,
        "SUB": sub,
        "DIV": div,
        "PD_LOC": pd_loc,
        "PD_EQ": pd_eq,
        "PD_MEAN": pd_mean,
        "PD_STD": pd_std,
        "PD_ANY": pd_any,
        "SK_FIT": sk_fit,
    }
    if external_op not in implemented_ops:
        raise NotImplementedError(
            f"{external_op} not in {list(implemented_ops.keys())}"
        )

    args = pkl.loads(transform_spec.external.arguments)
    kwargs = pkl.loads(transform_spec.external.named_arguments)
    func = implemented_ops[external_op]
    return await func(dataspec, *args, **kwargs)  # type: ignore



from __future__ import annotations

from datetime import datetime
from typing import List, Mapping, Optional, Type, cast

from sarus_data_spec.base import Referring
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.typing import Manager
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Status(Referring[sp.Status]):
    """A python class to describe status"""

    def __init__(self, protobuf: sp.Status) -> None:
        self._referred = {
            protobuf.dataset,
            protobuf.manager,
        }  # This has to be defined before it is initialized
        super().__init__(protobuf)

    def prototype(self) -> Type[sp.Status]:
        """Return the type of the underlying protobuf."""
        return sp.Status

    def datetime(self) -> datetime:
        return datetime.fromisoformat(self.protobuf().datetime)

    def ready(self) -> bool:
        return self.protobuf().stage.HasField("ready")

    def pending(self) -> bool:
        return self.protobuf().stage.HasField("pending")

    def processing(self) -> bool:
        return self.protobuf().stage.HasField("processing")

    def error(self) -> bool:
        return self.protobuf().stage.HasField("error")

    def dataset(self) -> Dataset:
        return cast(Dataset, self.storage().referrable(self._protobuf.dataset))

    def owner(
        self,
    ) -> Manager:  # TODO: Maybe find a better name, but this was shadowing the actual manager of this object.  # noqa: E501
        return cast(Manager, self.storage().referrable(self._protobuf.manager))


# Builders
def status(
    dataset: st.DataSpec,
    manager: Manager,
    stage: sp.Status.Stage,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
) -> Status:
    """A builder to ease the construction of a status"""
    # If none of fields or type is defined, set type to Null
    return Status(
        sp.Status(
            dataset=dataset.uuid(),
            manager=manager.uuid(),
            datetime=datetime.now().isoformat(),
            task=task,
            stage=stage,
            properties=properties,
        )
    )


def last_status(dataspec: st.DataSpec) -> Optional[st.Status]:
    """Return a DataSpec's last status by sorted datetime."""
    statuses = cast(
        List[st.Status], dataspec.referring(sp.type_name(sp.Status))
    )
    sorted_statuses = sorted(
        statuses, key=lambda x: x.datetime(), reverse=True
    )
    return next(iter(sorted_statuses), None)


def pending(
    dataset: st.DataSpec,
    manager: Manager,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
) -> Status:
    return status(
        dataset,
        manager,
        sp.Status.Stage(pending=sp.Status.Stage.Pending()),
        task,
        properties,
    )


def processing(
    dataset: st.DataSpec,
    manager: Manager,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
) -> Status:
    return status(
        dataset,
        manager,
        sp.Status.Stage(processing=sp.Status.Stage.Processing()),
        task,
        properties,
    )


def ready(
    dataset: st.DataSpec,
    manager: Manager,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
) -> Status:
    return status(
        dataset,
        manager,
        sp.Status.Stage(ready=sp.Status.Stage.Ready()),
        task,
        properties,
    )


def error(
    dataset: st.DataSpec,
    manager: Manager,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
) -> Status:
    return status(
        dataset,
        manager,
        sp.Status.Stage(error=sp.Status.Stage.Error()),
        task,
        properties,
    )



from __future__ import annotations

from typing import AsyncIterator, Iterator
import asyncio
import logging

import pandas as pd
import pyarrow as pa

from sarus_data_spec.manager.base import Base
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # error message printed from typing.py

from .utils import iter_over_async

logger = logging.getLogger(__name__)


class BaseAsyncManager(Base):
    """Asynchronous Manager Base implementation.

    Make synchronous methods rely on asynchronous ones for consistency.
    """

    def __init__(self, storage: Storage, protobuf: sp.Manager):
        super().__init__(
            storage=storage,
            protobuf=protobuf,
        )

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        raise NotImplementedError("async_value")

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        return asyncio.run(self.async_value(scalar=scalar))

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError("async_to_arrow")

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> Iterator[pa.RecordBatch]:
        batches_async_iterator = asyncio.run(
            self.async_to_arrow(dataset=dataset, batch_size=batch_size)
        )
        return iter_over_async(batches_async_iterator)

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        batches_async_it = await self.async_to_arrow(
            dataset=dataset, batch_size=64
        )
        arrow_batches = [batch async for batch in batches_async_it]
        return pa.Table.from_batches(arrow_batches).to_pandas()

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        return asyncio.run(self.async_to_pandas(dataset=dataset))

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        raise NotImplementedError("to_tensorflow")



from types import TracebackType
import typing as t

from sarus_data_spec.bounds import Bounds
from sarus_data_spec.context.state import (
    pop_global_context,
    push_global_context,
)
from sarus_data_spec.context.typing import Context
from sarus_data_spec.factory import Factory
from sarus_data_spec.manager.typing import Manager
from sarus_data_spec.marginals import Marginals
from sarus_data_spec.predicate import Predicate
from sarus_data_spec.protobuf import type_name
from sarus_data_spec.schema import Schema
from sarus_data_spec.size import Size
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec as s
import sarus_data_spec.protobuf as sp


class Base(Context):
    """A factory class with all the config"""

    def __init__(self) -> None:
        self._factory = Factory()
        # Register relevant classes
        self.factory().register(
            type_name(sp.Dataset),
            lambda protobuf: s.Dataset(t.cast(sp.Dataset, protobuf)),
        )
        self.factory().register(
            type_name(sp.Scalar),
            lambda protobuf: s.Scalar(t.cast(sp.Scalar, protobuf)),
        )
        self.factory().register(
            type_name(sp.Status),
            lambda protobuf: s.Status(t.cast(sp.Status, protobuf)),
        )
        self.factory().register(
            type_name(sp.Transform),
            lambda protobuf: s.Transform(t.cast(sp.Transform, protobuf)),
        )
        self.factory().register(
            type_name(sp.Attribute),
            lambda protobuf: s.Attribute(t.cast(sp.Attribute, protobuf)),
        )

        self.factory().register(
            type_name(sp.VariantConstraint),
            lambda protobuf: s.VariantConstraint(
                t.cast(sp.VariantConstraint, protobuf)
            ),
        )
        self.factory().register(
            type_name(sp.Predicate),
            lambda protobuf: Predicate(t.cast(sp.Predicate, protobuf)),
        )
        self.factory().register(
            type_name(sp.Schema),
            lambda protobuf: Schema(t.cast(sp.Schema, protobuf)),
        )
        self.factory().register(
            type_name(sp.Marginals),
            lambda protobuf: Marginals(t.cast(sp.Marginals, protobuf)),
        )
        self.factory().register(
            type_name(sp.Size),
            lambda protobuf: Size(t.cast(sp.Size, protobuf)),
        )
        self.factory().register(
            type_name(sp.Bounds),
            lambda protobuf: Bounds(t.cast(sp.Bounds, protobuf)),
        )

    def factory(self) -> Factory:
        return self._factory

    def storage(self) -> Storage:
        raise NotImplementedError()

    def manager(self) -> Manager:
        raise NotImplementedError()

    def __enter__(self) -> Context:
        push_global_context(self)
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        traceback: t.Optional[TracebackType],
    ) -> None:
        pop_global_context()
        return None



from itertools import count
from typing import Any, Generic, Tuple, TypeVar
import asyncio
import logging

from sarus_data_spec.context import global_context
from sarus_data_spec.status import error, last_status, processing, ready
import sarus_data_spec.typing as st


class DataSpecErrorStatus(Exception):
    ...


T = TypeVar("T")

logger = logging.getLogger(__name__)


class StatusAwareComputation(Generic[T]):
    """Singleton visitor class template to perform basic status checks.

    This class' purpose if to manage statuses, before, during and after a
    DataSpec's computation. It also manages caching.

    This class is intended to be subclassed and used in a Manager's
    `async_to_XXX` method. For instance, to implement `Manager.async_to_arrow`
    the idea is to implement a subclass ArrowComputation(Iterator[pa.
    RecordBatch]) and to do ArrowComputation.accept(dataspec).

    Subclasses must only implement the __read_from_cache__, __write_to_cache__
    and __process__ methods.
    """

    @classmethod
    async def accept(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """Apply the appropriate method depending on the status."""
        status = last_status(dataspec)

        if not status:
            return await cls.process(dataspec, *args, **kwargs)
        elif status.ready():
            return await cls.ready(dataspec, status)
        elif status.pending():
            return await cls.pending(dataspec, *args, **kwargs)
        elif status.processing():
            return await cls.processing(
                dataspec, *args, interval=20, attempts=10, **kwargs
            )
        elif status.error():
            return await cls.error(dataspec, status)
        else:
            raise ValueError(f"Inconsistent status {status}")

    @classmethod
    async def processing(
        cls,
        dataspec: st.DataSpec,
        *args: Any,
        attempts: int = 10,
        interval: int = 20,
        **kwargs: Any,
    ) -> T:
        """If processing, wait for the dataspec to be ready.

        Such a case can happen if another manager has launched the computation
        of this DataSpec's value. A timeout period is defined after which the
        computation is relaunched (this is a default behavior intended to be
        improved in the future).

        Args:
            attempts (int): number of attempts before timing out.
            interval (int): number of seconds between attempts.
        """
        attempt = count()
        while next(attempt) <= attempts:
            status = last_status(dataspec)
            assert status
            if not status.processing():
                break
            await asyncio.sleep(interval)

        status = last_status(dataspec)
        assert status
        if status.processing():
            # Still processing, process anyway
            # TODO might not be a good idea to relaunch
            # a really long computation
            return await cls.process(dataspec, *args, **kwargs)
        else:
            # computation finished, revisit to define what to do
            return await cls.accept(*args, **kwargs)

    @classmethod
    async def error(cls, dataspec: st.DataSpec, status: st.Status) -> T:
        """The DataSpec already has an Error status.

        In this case we read the error message stored in the status and reraise
        the exception with this message."""
        error_msg = status.properties()["message"]
        raise DataSpecErrorStatus(error_msg)

    @classmethod
    async def pending(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """The DataSpec has been added to the computation queue.

        No manager has undertaken the computation yet."""
        return await cls.process(dataspec, *args, **kwargs)

    @classmethod
    async def process(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """Compute the DataSpec's value.

        Set the DataSpec's status to PROCESSING and launch the computation
        defined in the subclass.

        If an error occurs, set the status to ERROR with the Exception
        representation.

        If no error occurs, write the result to cache, set the status to READY
        with the cache path and return the actual value.
        """
        manager = global_context().manager()
        processing(
            dataset=dataspec,
            manager=manager,
            task=cls.__name__,
        )
        try:
            result = await cls.__process__(dataspec, *args, **kwargs)
        except Exception as e:
            error(
                dataset=dataspec,
                manager=manager,
                task=cls.__name__,
                properties={"message": repr(e)},
            )
            logger.exception(e)
            raise DataSpecErrorStatus(repr(e))
        else:
            cache_path, result = await cls.__write_to_cache__(dataspec, result)
            ready(
                dataset=dataspec,
                manager=manager,
                task=cls.__name__,
                properties={"cache": cache_path},
            )
            return result

    @classmethod
    def cache_path(cls, status: st.Status) -> str:
        assert status.ready()
        return status.properties()["cache"]

    @classmethod
    async def ready(cls, dataspec: st.DataSpec, status: st.Status) -> T:
        """Read the value from the cache."""
        return await cls.__read_from_cache__(dataspec, status)

    @classmethod
    async def __read_from_cache__(
        cls, dataspec: st.DataSpec, status: st.Status
    ) -> T:
        # Should be implemented for each to_XXX method
        raise NotImplementedError

    @classmethod
    async def __write_to_cache__(
        cls, dataspec: st.DataSpec, data: T
    ) -> Tuple[str, T]:
        # Should be implemented for each to_XXX method
        raise NotImplementedError

    @classmethod
    async def __process__(
        cls,
        dataspec: st.DataSpec,
        *args: Any,
        **kwargs: Any,
    ) -> T:
        # Must be reimplemented for each to_XXX method
        raise NotImplementedError



from typing import Any

from .utils import two_arguments_ops


@two_arguments_ops
async def add(val_1: Any, val_2: Any) -> Any:
    return val_1 + val_2


@two_arguments_ops
async def sub(val_1: Any, val_2: Any) -> Any:
    return val_1 - val_2


@two_arguments_ops
async def mul(val_1: Any, val_2: Any) -> Any:
    return val_1 * val_2


@two_arguments_ops
async def div(val_1: Any, val_2: Any) -> Any:
    return val_1 / val_2



from __future__ import annotations

from typing import Any, cast

try:
    import sklearn
except ModuleNotFoundError:
    pass  # error message in typing.py

import sarus_data_spec.typing as st

from .utils import pandas_or_value


async def sk_fit(
    scalar: st.Scalar, *args: Any, **kwargs: Any
) -> sklearn.svm.SVC:
    (ds_model, X, y), _ = scalar.parents()
    model = await cast(st.Scalar, ds_model).async_value()
    assert isinstance(model, sklearn.svm.SVC)
    X = await pandas_or_value(X)
    y = await pandas_or_value(y)
    fitted_model = model.fit(X, y, *args, **kwargs)
    return fitted_model



from __future__ import annotations

from typing import Any, Dict
import pickle as pkl

try:
    import sklearn
except ModuleNotFoundError:
    pass  # error message in typing.py

import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


async def create_model(scalar: st.Scalar) -> sklearn.svm.SVC:
    model_spec = scalar.protobuf().spec.model

    args = pkl.loads(model_spec.arguments)
    kwargs = pkl.loads(model_spec.named_arguments)

    model_mapping: Dict[sp.Scalar.Model.ModelClass.V, Any] = {
        sp.Scalar.Model.ModelClass.SK_SVC: sklearn.svm.SVC,
    }
    ModelClass = model_mapping[model_spec.model_class]

    return ModelClass(*args, **kwargs)



from typing import Any, List, Tuple, Union
import logging

import pandas as pd

from .utils import one_parent_ops, two_arguments_ops

logger = logging.getLogger(__name__)


@one_parent_ops
async def pd_loc(
    parent_val: Any, key: Tuple[Union[str, slice, List[str]], ...]
) -> pd.DataFrame:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return pd.core.indexing._LocIndexer.__getitem__(parent_val.loc, key)


@two_arguments_ops
async def pd_eq(val_1: Any, val_2: Any) -> pd.DataFrame:
    return val_1 == val_2


@one_parent_ops
async def pd_mean(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.mean(*args, **kwargs)


@one_parent_ops
async def pd_std(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.std(*args, **kwargs)


@one_parent_ops
async def pd_any(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.any(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/scalar.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/scalar.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%sarus_data_spec/protobuf/scalar.proto\x12\x0fsarus_data_spec\"\xa5\x05\n\x06Scalar\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12*\n\x04spec\x18\x04 \x01(\x0b\x32\x1c.sarus_data_spec.Scalar.Spec\x12;\n\nproperties\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Scalar.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1az\n\x04Spec\x12:\n\x0btransformed\x18\x01 \x01(\x0b\x32#.sarus_data_spec.Scalar.TransformedH\x00\x12.\n\x05model\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Scalar.ModelH\x00\x42\x06\n\x04spec\x1a\xbc\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12P\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x37.sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x9a\x01\n\x05Model\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12=\n\x0bmodel_class\x18\x03 \x01(\x0e\x32(.sarus_data_spec.Scalar.Model.ModelClass\"&\n\nModelClass\x12\x0c\n\x08TF_KERAS\x10\x00\x12\n\n\x06SK_SVC\x10\x01\x62\x06proto3'
)



_SCALAR_MODEL_MODELCLASS = _descriptor.EnumDescriptor(
  name='ModelClass',
  full_name='sarus_data_spec.Scalar.Model.ModelClass',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TF_KERAS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SK_SVC', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=698,
  serialized_end=736,
)
_sym_db.RegisterEnumDescriptor(_SCALAR_MODEL_MODELCLASS)


_SCALAR_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Scalar.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Scalar.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Scalar.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=215,
  serialized_end=264,
)

_SCALAR_SPEC = _descriptor.Descriptor(
  name='Spec',
  full_name='sarus_data_spec.Scalar.Spec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformed', full_name='sarus_data_spec.Scalar.Spec.transformed', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='sarus_data_spec.Scalar.Spec.model', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='spec', full_name='sarus_data_spec.Scalar.Spec.spec',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=266,
  serialized_end=388,
)

_SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY = _descriptor.Descriptor(
  name='NamedArgumentsEntry',
  full_name='sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=526,
  serialized_end=579,
)

_SCALAR_TRANSFORMED = _descriptor.Descriptor(
  name='Transformed',
  full_name='sarus_data_spec.Scalar.Transformed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transform', full_name='sarus_data_spec.Scalar.Transformed.transform', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='sarus_data_spec.Scalar.Transformed.arguments', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_arguments', full_name='sarus_data_spec.Scalar.Transformed.named_arguments', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=391,
  serialized_end=579,
)

_SCALAR_MODEL = _descriptor.Descriptor(
  name='Model',
  full_name='sarus_data_spec.Scalar.Model',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='arguments', full_name='sarus_data_spec.Scalar.Model.arguments', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_arguments', full_name='sarus_data_spec.Scalar.Model.named_arguments', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_class', full_name='sarus_data_spec.Scalar.Model.model_class', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SCALAR_MODEL_MODELCLASS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=582,
  serialized_end=736,
)

_SCALAR = _descriptor.Descriptor(
  name='Scalar',
  full_name='sarus_data_spec.Scalar',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='sarus_data_spec.Scalar.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Scalar.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='doc', full_name='sarus_data_spec.Scalar.doc', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spec', full_name='sarus_data_spec.Scalar.spec', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Scalar.properties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SCALAR_PROPERTIESENTRY, _SCALAR_SPEC, _SCALAR_TRANSFORMED, _SCALAR_MODEL, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=59,
  serialized_end=736,
)

_SCALAR_PROPERTIESENTRY.containing_type = _SCALAR
_SCALAR_SPEC.fields_by_name['transformed'].message_type = _SCALAR_TRANSFORMED
_SCALAR_SPEC.fields_by_name['model'].message_type = _SCALAR_MODEL
_SCALAR_SPEC.containing_type = _SCALAR
_SCALAR_SPEC.oneofs_by_name['spec'].fields.append(
  _SCALAR_SPEC.fields_by_name['transformed'])
_SCALAR_SPEC.fields_by_name['transformed'].containing_oneof = _SCALAR_SPEC.oneofs_by_name['spec']
_SCALAR_SPEC.oneofs_by_name['spec'].fields.append(
  _SCALAR_SPEC.fields_by_name['model'])
_SCALAR_SPEC.fields_by_name['model'].containing_oneof = _SCALAR_SPEC.oneofs_by_name['spec']
_SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY.containing_type = _SCALAR_TRANSFORMED
_SCALAR_TRANSFORMED.fields_by_name['named_arguments'].message_type = _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY
_SCALAR_TRANSFORMED.containing_type = _SCALAR
_SCALAR_MODEL.fields_by_name['model_class'].enum_type = _SCALAR_MODEL_MODELCLASS
_SCALAR_MODEL.containing_type = _SCALAR
_SCALAR_MODEL_MODELCLASS.containing_type = _SCALAR_MODEL
_SCALAR.fields_by_name['spec'].message_type = _SCALAR_SPEC
_SCALAR.fields_by_name['properties'].message_type = _SCALAR_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['Scalar'] = _SCALAR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Scalar = _reflection.GeneratedProtocolMessageType('Scalar', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _SCALAR_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar.PropertiesEntry)
    })
  ,

  'Spec' : _reflection.GeneratedProtocolMessageType('Spec', (_message.Message,), {
    'DESCRIPTOR' : _SCALAR_SPEC,
    '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar.Spec)
    })
  ,

  'Transformed' : _reflection.GeneratedProtocolMessageType('Transformed', (_message.Message,), {

    'NamedArgumentsEntry' : _reflection.GeneratedProtocolMessageType('NamedArgumentsEntry', (_message.Message,), {
      'DESCRIPTOR' : _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY,
      '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry)
      })
    ,
    'DESCRIPTOR' : _SCALAR_TRANSFORMED,
    '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar.Transformed)
    })
  ,

  'Model' : _reflection.GeneratedProtocolMessageType('Model', (_message.Message,), {
    'DESCRIPTOR' : _SCALAR_MODEL,
    '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar.Model)
    })
  ,
  'DESCRIPTOR' : _SCALAR,
  '__module__' : 'sarus_data_spec.protobuf.scalar_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Scalar)
  })
_sym_db.RegisterMessage(Scalar)
_sym_db.RegisterMessage(Scalar.PropertiesEntry)
_sym_db.RegisterMessage(Scalar.Spec)
_sym_db.RegisterMessage(Scalar.Transformed)
_sym_db.RegisterMessage(Scalar.Transformed.NamedArgumentsEntry)
_sym_db.RegisterMessage(Scalar.Model)


_SCALAR_PROPERTIESENTRY._options = None
_SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
# @@protoc_insertion_point(module_scope)
